'''Safe execution of arbitrary code through bytecode munging

CodeBlock is a base class for bytecode munging of code strings.  Derived
classes should override 'forbid' and 'Mungers'.  The class constructor
takes a function name, a parameter string, and a function body string. It
combines the name, parameters, and body into a function (indented one
space), compiles it, and then examines its bytecode.

Any bytecode whose opcode is a key in 'forbid' with a true value may not
appear in the compiled code.

Each object in 'Mungers' is called with the code window object (at start)
and the variable name information for the function.  Mungers can use this
opportunity to examine the code and variables, and perform setup operations.
The value returned from the call is discarded if false, otherwise it is
assumed to be a collection of bytecode-specific functions.

The code window object is passed over the bytecodes, maintaining a stack
of "responsible" opcodes.  At any given position, the size of the stack
should be the same as the size of the run-time stack would be when that
part of the code is executing, and the top of the stack would have been
put there by the "responsible" opcode.

At each position, the mungers are examined to see if they have a function
corresponding to the current opcode.  If so, the function is called with
the code window object.  The function can then use the code window object to
examine or change the bytecode.

Once all processing is done, and if there have not been any errors, the
compiled, edited function is converted into a pickleable tuple, which is
stored as attribute 't' of the resulting instance.  In order for this to
work, mungers must not place unpickleable objects into the list of constants.
In order for unpickling to be robust, they should not place any function
objects there either.  If a function must be provided for use in the
bytecode, it should be loaded from the global dictionary, preferably with an
'illegal' name.
'''

from string import replace, join
import ops, code_editor
from closure import bind

class Warning(Exception): pass
class ForbiddenOp(Exception): pass

class Munge_window:
    def __init__(self, code, use_stack):
        self.code = code
        self.opn = -1
        self.use_stack = use_stack
        if use_stack:
            self.stack = []
            self.last_empty_stack = None
    def insert_code(self, rcode, at = None):
        opn = self.opn
        # rcode is passed as a (callables, args) pair
        rcode = map(apply, rcode[0], rcode[1])
        if at is None:
            at = opn
        self.code[at:at]=rcode
        self.opn = opn + len(rcode)
        return rcode

    def set_code(self, rct = 1, rcode = ()):
        'Replace "rct" bytecodes at current position with "rcode"'
        opn = self.opn
        # rcode is passed as a (callables, args) pair
        if rcode:
            rcode = map(apply, rcode[0], rcode[1])
        self.code[opn:opn+rct] = rcode
        if self.use_stack:
            stack = self.stack
            for op in rcode:
                try:
                    op.execute(stack)
                except:
                    del stack[:]
                if not stack:
                    self.last_empty_stack = op
        self.opn = opn + len(rcode) - 1
    def advance(self):
        self.opn = opn = self.opn + 1
        code = self.code
        if opn < len(code):
            self.op = code.opcodes[opn]
            return 1
    def do_op(self):
        if self.use_stack:
            op = self.op
            stack = self.stack
            try:
                op.execute(stack)
            except:
                del stack[:]
            if not stack:
                self.last_empty_stack = op
    def after_code_for(self, stack_idx):
        stack = self.stack
        if stack_idx < 0 and len(stack) + stack_idx < 0:
            whichop = self.last_empty_stack
            if whichop is None:
                return 0
        else:
            whichop = stack[stack_idx]
        return self.code.index(whichop)+1

    def assert_stack_size(self, size, ctxt):
        if self.use_stack and len(self.stack) != size:
            raise Warning, ('PythonMethod Bug: %s objects were on the stack '
            '%s when only %s should be.' % (len(self.stack), ctxt, size)) 

class CodeBlock:
    '''Compile a string containing Python code, with restrictions'''
    def __init__(self, name, paramstr, src):
        self.name = name
        self.src = src
        self.paramstr = paramstr
        self.f = None
        defns = {'__builtins__': None}
        defblk = 'def %s(%s):\n %s\n' % (name, self.paramstr
        , replace(self.src, '\n', '\n '))
        exec defblk in defns
        block = defns[name]
        self.munge(block)

    forbid = {}
    Mungers = ()
    globals = {'__builtins__': None}

    def munge(self, block):
	f = code_editor.Function(block)
        f.func_globals = self.globals
	fc =f.func_code
	code = fc.co_code
        self.warnings, self.errors = warnings, errors = [], []
	mungers = []
        window = Munge_window(code, 1)
	margs = (window, fc.co_varnames, fc.co_names, fc.co_consts, warnings)
	for M in self.Mungers:
            try:
                mungers.append(apply(M, margs))
            except Exception, e:
                errors.append(e.__class__.__name__ + ', ' + str(e))
        if errors:
            return
        mungers = filter(None, mungers)
        line = 0 ; forbid = self.forbid
        while window.advance():
            op = window.op
            if isinstance(op, ops.SET_LINENO):
                line = op.arg
                window.do_op()
            elif forbid.get(op.op, 0):
                errors.append("Forbidden operation %s at line %d" % (op.__class__.__name__, line))
            else:
                for m in mungers:
                    handler = getattr(m, op.__class__.__name__, None)
                    if handler:
                        try:
                            handled = handler(window)
                        except ForbiddenOp:
                            errors.append("Forbidden operation %s at line %d" % (op.__class__.__name__, line))
                        except Exception, e:
                            raise
                            errors.append(e.__class__.__name__ + ', ' + str(e))
                        else:
                            if not handled:
                                continue
                        break
                else:
                    window.do_op()
        if not errors:
            self.t = f.as_tuple()
    def __call__(self, *args, **kargs):
        F = code_editor.Function(self.t)
        F.func_globals = self.globals
        self.__call__ = f = F.make_function()
        return apply(f, args, kargs)

def _print_handler(printlist, *txt):
    add = printlist.append
    if len(txt):
        if printlist and printlist[-1:] != ['\n']:
            add(' ')
        add(str(txt[0]))
    else:
        add('\n')
def _join_printed(printlist):
    return join(printlist, '')
_join_printed = bind(_join_printed, join=join, map=map, str=str)

class Printing:
    '''Intercept print statements

    Print statements are either converted to no-ops, or replaced with
    calls to a handler.  The default handler _print_handler appends the
    intended output to a list.  'printed' is made into a reserved name
    which can only be used to read the result of _join_printed.

    _print_handler and _join_printed should be provided in the global
    dictionary as $print_handler and $join_printed'''
    discard = ((ops.POP_TOP,), ((),))
    lfnames = (('$printed',), ('$print_handler',), ('$join_printed',))
    make_printed = ( (ops.BUILD_LIST, ops.STORE_FAST,
                      ops.LOAD_GLOBAL, ops.STORE_FAST,
                      ops.LOAD_GLOBAL, ops.STORE_FAST),
                     ((0,), lfnames[0],
                      lfnames[1], lfnames[1],
                      lfnames[2], lfnames[2]) )
    call_print1 = ( (ops.CALL_FUNCTION, ops.POP_TOP), ((1,), ()) )
    call_print2 = ( (ops.CALL_FUNCTION, ops.POP_TOP), ((2,), ()) )
    
    def __init__(self, w, varnames, names, consts, warnings):
        self.warnings = warnings
        if 'printed' in names:
            lfns = self.lfnames
            self.pidx = names.index('printed')
            self.print_prep = ((ops.LOAD_FAST, ops.LOAD_FAST),
                               (lfns[1], lfns[0]) )
            self.get_printed = (
                (ops.LOAD_FAST, ops.LOAD_FAST, ops.CALL_FUNCTION),
                (lfns[2], lfns[0], (1,)) )
            # At the start of the function, set up the printing framework.
            # Don't bother with the op stack, since no code can bracket this.
            w.insert_code(self.make_printed, 0)
        else:
            self.PRINT_ITEM = self.PRINT_NEWLINE = self.dev_null
            self.LOAD_GLOBAL = None
            pops = (ops.PRINT_ITEM.op, ops.PRINT_NEWLINE.op)
            for op in w.code:
                if op.op in pops:
                    warnings.append("Prints, but never reads 'printed' variable.")
                    break
        if 'printed' in varnames:
            raise SyntaxError, '"printed" is a reserved word.'
    def dev_null(self, w):
        if isinstance(w.op, ops.PRINT_NEWLINE):
            w.set_code()
        else:
            w.set_code(1, self.discard)
        return 1
    def PRINT_ITEM(self, w):
        # Load the printing function before the code for the operand.
        w.insert_code(self.print_prep, w.after_code_for(-2))
        w.assert_stack_size(1, "at a 'print' statement")
        # Instead of printing, call our function and discard the result.
        w.set_code(1, self.call_print2)
        return 1
    def PRINT_NEWLINE(self, w):
        w.assert_stack_size(0, "at a 'print' statement")
        w.insert_code(self.print_prep)
        w.set_code(1, self.call_print1)
        return 1
    def LOAD_GLOBAL(self, w):
        if w.op.arg==self.pidx:
            # Construct the print result instead of getting non-existent var.
            w.set_code(1, self.get_printed)
            return 1

def PublicNames(w, varnames, names, consts, warnings):
    '''Restrict access to all but public names

    Forbid use of any multi-character name starting with _
    '''
    protected = []
    for name in names:
        if name[:1]=='_' and len(name)>1:
            protected.append(name)
    if protected:
        raise SyntaxError, ('Names starting with "_" are not allowed (%s).'
                            % join(protected, ', '))

class AllowMapBuild:
    '''Allow literal mappings

    Construction of literal dictionaries requires STORE_SUBSCR, which is
    usually forbidden.  Allow it by checking the stack.'''
    def __init__(self, w, varnames, names, consts, warnings):
        pass
    def STORE_SUBSCR(self, w):
        if isinstance(w.stack[-2], ops.BUILD_MAP):
            return
        raise ForbiddenOp

def _get_call(w):
    load_guard = ((ops.LOAD_FAST, ops.LOAD_ATTR), (('$guard',), (guard,)))
    # Load the binary guard function before its parameters are computed.
    iops = w.insert_code(load_guard, w.after_code_for(-3))
    # Fix the execution stack to refer to the loaded function.
    if w.use_stack: w.stack[-2:-2] = iops[1:]
    # Call guard function instead of performing binary op
    w.set_code(1, cf2)
    return 1
def _SLICE(w):
    load_guard = ((ops.LOAD_FAST, ops.LOAD_ATTR, ops.ROT_TWO),
                  (('$guard',), ('getslice',), ()))
    # Load the Slice guard and switch it with the argument.
    w.insert_code(load_guard, w.opn+1)
    # Call the slice guard after performing the slice.
    w.insert_code(cf1, w.opn+4)
    return 1

class _GuardedOps:
    def __call__(self, w, varnames, names, consts, warnings):
        localize_guard = ( (ops.LOAD_GLOBAL, ops.STORE_FAST), (('$guard',),('$guard',)) )
        # Insert setup code; no need to fix stack
        w.insert_code(localize_guard, 0)
        return self

def GuardedOps(guards):
    '''Allow operations to be Guarded, by replacing them with guard calls.

    Construct a munging object which will replace specific opcodes with
    calls to methods of a guard object.  The guard object must appear in
    the global dictionary as $guard.'''
    gops = _GuardedOps()
    opmap = (('mul', 'BINARY_MULTIPLY')
            ,('div', 'BINARY_DIVIDE')
            ,('power', 'BINARY_POWER')
            ,('getattr', 'LOAD_ATTR')
            ,('getitem', 'BINARY_SUBSCR'))
    cf1 = ((ops.CALL_FUNCTION,), ((1,),))
    cf2 = ((ops.CALL_FUNCTION,), ((2,),))
    for guard, defname in opmap:
        if hasattr(guards, guard):
            setattr(gops, defname, bind(_get_call, guard=guard, cf2=cf2))
    if hasattr(guards, 'getslice'):
        gops.SLICE_3 = gops.SLICE_2 = gops.SLICE_1 = gops.SLICE_0 = bind(_SLICE, cf1=cf1)
    return gops

class SafeBlock(CodeBlock):
    forbid = {ops.STORE_FAST.op: 0, ops.STORE_GLOBAL.op: 0, ops.STORE_SUBSCR.op: 0}
    for opclass in ops._bytecodes.values():
        if forbid.get(opclass.op, 1):
            opname = opclass.__name__
            if (opname[:6]=='STORE_' or opname[:7]=='DELETE_' or opname[:5]=='EXEC_'):
                forbid[opclass.op] = 1
    Mungers = [Printing, PublicNames, AllowMapBuild]

def UntupleFunction(t, special_globals, **globals):
    t = list(t)
    t[-1] = globals
    globals.update(special_globals)
    f = code_editor.Function(t)
    fc = f.func_code
    fcn, fcvn = fc.co_names, fc.co_varnames
    for name, _f in (('$print_handler', _print_handler),
                     ('$join_printed', _join_printed)):
        if name in fcn:
            globals[name] = _f
    defined = globals.has_key
    if not defined('__builtins__'):
        globals['__builtins__'] = {}
    globals['global_exists'] = globals.has_key
    gbi = globals['__builtins__'].has_key
    #for name in fcn:
    #    if name in fcvn or name[:1]=='$' or defined(name) or gbi(name):
    #        pass
    #    else:
    #        globals[name] = None
    return f.make_function()

if 0:
    #demonstrator Guard
    class Guard:
        def mul(self, a, b):
            if a*b>10:
                return 'ha!'
            else:
                return a*b
        def getslice(self, s):
            if len(s)>2:
                raise "Too long a slice!"
            return s

def test(p, c):
    sb = SafeBlock('f', p, c)
    print sb.errors, sb.warnings
    f = code_editor.Function(sb.t)
    for c in f.func_code.co_code:
        print c
    return sb

if __name__ == '__main__':    
	sb = test('x', 'print x\nreturn printed')
	#from dis import dis
	#dis(sb.block)
	print sb(2),
	print sb(3),

